from typing import Any, Dict, Optional

from asyncio import ensure_future
from dataclasses import asdict, dataclass

from aiohttp import web
from aiohttp.web import Application, AppRunner, TCPSite, RouteTableDef, \
    Request, Response
from aiohttp.helpers import sentinel
from aiohttp.typedefs import LooseHeaders
from dazl.protocols.v0.json_ser_command import LedgerJSONEncoder


from .log import LOG
from .config import Configuration
from .integration_context import IntegrationContext

# cap aiohttp to allow a maximum of 100 MB for the size of a body.
CLIENT_MAX_SIZE = 100 * (1024 ** 2)

DEFAULT_ENCODER = LedgerJSONEncoder()


def json_response(
        data: Any = sentinel, *,
        text: str = None,
        body: bytes = None,
        status: int = 200,
        reason: 'Optional[str]' = None,
        headers: 'LooseHeaders' = None) -> 'web.Response':
    return web.json_response(
        data=data, text=text, body=body, status=status, reason=reason, headers=headers,
        dumps=lambda obj: DEFAULT_ENCODER.encode(obj) + '\n')


def unauthorized_response(code: str, description: str) -> 'web.HTTPUnauthorized':
    body = DEFAULT_ENCODER.encode({'code': code, 'description': description}) + '\n'
    return web.HTTPUnauthorized(text=body, content_type='application/json')


def forbidden_response(code: str, description: str) -> 'web.HTTPForbidden':
    body = DEFAULT_ENCODER.encode({'code': code, 'description': description}) + '\n'
    return web.HTTPForbidden(text=body, content_type='application/json')


def not_found_response(code: str, description: str) -> 'web.HTTPNotFound':
    body = DEFAULT_ENCODER.encode({'code': code, 'description': description}) + '\n'
    return web.HTTPNotFound(text=body, content_type='application/json')


def bad_request(code: str, description: str) -> 'web.HTTPBadRequest':
    body = DEFAULT_ENCODER.encode({'code': code, 'description': description}) + '\n'
    return web.HTTPBadRequest(text=body, content_type='application/json')


def internal_server_error(code: str, description: str) -> 'web.HTTPInternalServerError':
    body = DEFAULT_ENCODER.encode({'code': code, 'description': description}) + '\n'
    return web.HTTPInternalServerError(text=body, content_type='application/json')


def _build_healthcheck_route(
        integration_context: 'IntegrationContext') -> 'RouteTableDef':
    routes = RouteTableDef()

    @routes.get('/healthz')
    async def get_container_health(request: 'Request') -> 'Response':
        response_dict = {
            '_self': str(request.url),
            'integration': asdict(integration_context.get_status())
        }
        return json_response(response_dict)

    return routes


async def start_web_endpoint(
        config: 'Configuration',
        integration_context: 'IntegrationContext'):

    # prepare the web application
    app = Application(client_max_size=CLIENT_MAX_SIZE)

    app.add_routes(_build_healthcheck_route(integration_context))

    if integration_context.running and integration_context.webhook_context:
        app.add_routes(integration_context.webhook_context.route_table)

    LOG.info('Opening a TCP socket...')
    runner = AppRunner(app, access_log_format='%a %t "%r" %s %b')
    await runner.setup()
    site = TCPSite(runner, '0.0.0.0', config.health_port)

    LOG.info('Started the web server on port %s.', config.health_port)
    return ensure_future(site.start())



import sys
import collections

from dataclasses import dataclass
from functools import wraps
from typing import Optional

from dazl import AIOPartyClient

from daml_dit_api import IntegrationResponse

from .log import LOG


@dataclass
class InvocationStatus:
    index: int
    command_count: int
    use_count: int
    error_count: int
    error_message: 'Optional[str]'


def without_return_value(fn):

    @wraps(fn)
    async def wrapped(*args, **kwargs):
        await fn(*args, **kwargs)

    return wrapped


def with_marshalling(mfn, fn):

    @wraps(fn)
    async def wrapped(arg):
        await fn(mfn(arg))

    return wrapped


def normalize_integration_response(response):
    LOG.info('Normalizing integration response: %r', response)

    if isinstance(response, IntegrationResponse):
        LOG.info('Integration Response passthrough')
        return response

    commands = []
    if isinstance(response, collections.Sequence):
        commands = response
    elif response:
        commands = [response]
    else:
        commands = []

    LOG.info('Integration response with cmds: %r', commands)

    return IntegrationResponse(commands=commands)


def as_handler_invocation(client: 'AIOPartyClient', inv_status: 'InvocationStatus', fn):
    @wraps(fn)
    async def wrapped(*args, **kwargs):
        LOG.info('Invoking for invocation status: %r', inv_status)
        inv_status.use_count += 1

        response = None
        try:
            response = normalize_integration_response(
                await fn(*args, **kwargs))

            if response.commands:
                LOG.debug('Submitting ledger commands: %r', response.commands)

                inv_status.command_count += len(response.commands)
                await client.submit(response.commands)

            return response

        except Exception:
            inv_status.error_count += 1
            inv_status.error_message = repr(sys.exc_info()[1])
            LOG.exception('Error while processing: ' + inv_status)

    return wrapped



from dataclasses import dataclass
from typing import Optional, Sequence
from functools import wraps

from aiohttp import web
from aiohttp.web import RouteTableDef

from .common import \
    InvocationStatus, \
    without_return_value, \
    as_handler_invocation

from daml_dit_api import IntegrationWebhookRoutes

from .log import LOG


def empty_success_response() -> 'web.HTTPOk':
    return web.HTTPOk()


@dataclass
class WebhookRouteStatus(InvocationStatus):
    url_path: str
    method: str
    label: 'Optional[str]'


@dataclass
class IntegrationWebhookStatus:
    routes: 'Sequence[WebhookRouteStatus]'


class IntegrationWebhookContext(IntegrationWebhookRoutes):

    def __init__(self, client):
        self.route_table = RouteTableDef()
        self.client = client

        self.routes = []  # type: List[WebhookRouteStatus]

    def _with_resp_handling(self, status: 'InvocationStatus', fn):

        @wraps(fn)
        async def wrapped(request):
            hook_response = await fn(request)

            if hook_response.response is None:
                response = empty_success_response()
            else:
                response = hook_response.response

            LOG.debug('Webhook Response: %r', response)

            return response

        return wrapped

    def _notice_hook_route(self, url_path: str, method: str,
                           label: 'Optional[str]') -> 'WebhookRouteStatus':

        route_status = \
            WebhookRouteStatus(
                index=len(self.routes),
                url_path=url_path,
                method=method,
                label=label,
                command_count=0,
                use_count=0,
                error_count=0,
                error_message=None)

        self.routes.append(route_status)

        return route_status

    def _url_path(self, url_suffix: 'Optional[str]'):
        return '/integration/{integration_id}' + (url_suffix or '')

    def post(self, url_suffix: 'Optional[str]' = None, label: 'Optional[str]' = None):
        path = self._url_path(url_suffix)
        hook_status = self._notice_hook_route(path, 'post', label)

        def wrap_method(func):
            return self.route_table.post(path=path)(
                self._with_resp_handling(
                    hook_status,
                    as_handler_invocation(
                        self.client, hook_status, func)))

        return wrap_method

    def get(self, url_suffix: 'Optional[str]' = None, label: 'Optional[str]' = None):
        path = self._url_path(url_suffix)
        hook_status = self._notice_hook_route(path, 'get', label)

        def wrap_method(func):
            return self.route_table.get(path=path)(
                self._with_resp_handling(
                    hook_status,
                    as_handler_invocation(
                        self.client, hook_status, func)))

        return wrap_method

    def get_status(self) -> 'IntegrationWebhookStatus':
        return IntegrationWebhookStatus(
            routes=self.routes)



import logging
import time


def setup_default_logging(**overrides):
    logging.Formatter.converter = time.gmtime

    defaults = {
        'level': logging.INFO,
        'format': '%(asctime)s [%(levelname)s] (%(name)s) %(message)s',
        'datefmt': '%Y-%m-%dT%H:%M:%S%z'
    }

    config = {**defaults, **overrides}

    logging.basicConfig(**config)


LOG = logging.getLogger('integration')



import asyncio
from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Sequence, Tuple

from dazl import AIOPartyClient
from dazl.model.core import ContractMatch
from dazl.model.reading import ContractCreateEvent
from dazl.model.writing import EventHandlerResponse

from daml_dit_api import \
    IntegrationLedgerEvents, \
    IntegrationLedgerContractCreateEvent, \
    IntegrationLedgerContractArchiveEvent, \
    IntegrationLedgerTransactionStartEvent, \
    IntegrationLedgerTransactionEndEvent

from .common import \
    InvocationStatus, \
    without_return_value, \
    as_handler_invocation, \
    with_marshalling

from .log import LOG


Sweep = Tuple[Any,
              Optional[ContractMatch],
              Callable[[ContractCreateEvent], EventHandlerResponse]]


PendingHandlerCall = Callable[[], None]


@dataclass
class LedgerHandlerStatus(InvocationStatus):
    description: str
    sweep_enabled: bool
    flow_enabled: bool


@dataclass
class IntegrationLedgerStatus:
    pending_calls: int
    sweep_events: int
    event_handlers: 'Sequence[LedgerHandlerStatus]'


class IntegrationLedgerContext(IntegrationLedgerEvents):
    def __init__(self, client: 'AIOPartyClient'):
        self.pending_queue = \
            asyncio.Queue()  # type: asyncio.Queue[PendingHandlerCall]

        self.client = client
        self.sweep_events = 0
        self.handlers = []  # type: List[LedgerHandlerStatus]
        self.sweeps = []  # type: List[Sweep]
        self.init_handlers = []  # type: List[PendingHandlerCall]
        self.ready_handlers = []  # type: List[PendingHandlerCall]

    def _notice_handler(
            self, description: str,
            sweep_enabled: bool, flow_enabled: bool) -> 'LedgerHandlerStatus':

        handler_status = \
            LedgerHandlerStatus(
                index=len(self.handlers),
                description=description,
                command_count=0,
                use_count=0,
                error_count=0,
                error_message=None,
                sweep_enabled=sweep_enabled,
                flow_enabled=flow_enabled)

        self.handlers.append(handler_status)

        return handler_status

    async def worker(self):
        LOG.info('Ledger context worker starting...')
        while True:
            LOG.info('...waiting for ledger event....')
            pending_call = await self.pending_queue.get()
            try:
                LOG.debug('...received ledger event: %r', pending_call)
                await pending_call()

            except:  # noqa: E722
                LOG.exception('Uncaught error in Ledger context worker loop.')

    async def process_sweeps(self):
        LOG.info("Invoking init handlers")

        for init_handler in self.init_handlers:
            await init_handler()

        for (template, match, wfunc) in self.sweeps:
            LOG.info('Processing sweep for %r', template)

            for (cid, cdata) in self.client.find_active(template, match).items():
                LOG.info('Sweep contract: %r => %r', cid, cdata)

                self.sweep_events += 1

                await wfunc(IntegrationLedgerContractCreateEvent(
                    initial=True,
                    cid=cid,
                    cdata=cdata))

        LOG.info("Sweeps processed, invoking ready handlers")

        for ready_handler in self.ready_handlers:
            await ready_handler()

        LOG.info("Done with ready handlers")

    def _with_deferral(self, handler):

        async def wrapped(*args, **kwargs):
            async def pending():
                await handler(*args, **kwargs)

            LOG.info("Deferring call: %r", pending)
            await self.pending_queue.put(pending)

        return wrapped

    def _to_int_create_event(self, dazl_event):
        return IntegrationLedgerContractCreateEvent(
            initial=False,
            cid=dazl_event.cid,
            cdata=dazl_event.cdata)

    def ledger_init(self):
        handler_status = self._notice_handler('Ledger Init', False, True)

        def wrap_method(func):
            handler = self._with_deferral(
                without_return_value(
                    as_handler_invocation(
                        self.client, handler_status, func)))

            self.init_handlers.append(handler)

            return handler

        return wrap_method

    def ledger_ready(self):
        handler_status = self._notice_handler('Ledger Ready', False, True)

        def wrap_method(func):
            handler = self._with_deferral(
                without_return_value(
                    as_handler_invocation(
                        self.client, handler_status, func)))

            self.ready_handlers.append(handler)

            return handler

        return wrap_method

    def transaction_start(self):
        handler_status = self._notice_handler('Transaction Start', False, True)

        def to_int_event(dazl_event):
            return IntegrationLedgerTransactionEndEvent(
                command_id=dazl_event.command_id,
                workflow_id=dazl_event.workflow_id,
                contract_events=[self._to_int_create_event(e)
                                 for e in dazl_event.contract_events])

        def wrap_method(func):
            handler = self._with_deferral(
                with_marshalling(
                    to_int_event,
                    without_return_value(
                        as_handler_invocation(
                            self.client, handler_status, func))))

            self.client.add_ledger_transaction_start(handler)

            return handler

        return wrap_method

    def transaction_end(self):
        handler_status = \
            self._notice_handler('Transaction End', False, True)

        def to_int_event(dazl_event):
            return IntegrationLedgerTransactionEndEvent(
                command_id=dazl_event.command_id,
                workflow_id=dazl_event.workflow_id,
                contract_events=[self._to_int_create_event(e)
                                 for e in dazl_event.contract_events])

        def wrap_method(func):
            handler = self._with_deferral(
                with_marshalling(
                    to_int_event,
                    without_return_value(
                        as_handler_invocation(
                            self.client, handler_status, func))))

            self.client.add_ledger_transaction_end(handler)

            return handler

        return wrap_method

    def contract_created(
            self, template: Any, match: 'Optional[ContractMatch]' = None,
            sweep: bool = True, flow: bool = True):

        handler_status = \
            self._notice_handler(f'Contract Create - {template}', sweep, flow)

        def wrap_method(func):
            wfunc = self._with_deferral(
                without_return_value(
                    as_handler_invocation(
                        self.client, handler_status, func)))

            if sweep:
                self.sweeps.append((template, match, wfunc))

            handler = with_marshalling(self._to_int_create_event, wfunc)

            if flow:
                self.client.add_ledger_created(template, match=match, handler=handler)

            return handler

        return wrap_method

    def contract_archived(self, template: Any, match: 'Optional[ContractMatch]' = None):

        handler_status = \
            self._notice_handler(f'Contract Archive - {template}', False, True)

        def to_int_event(dazl_event):
            return IntegrationLedgerContractArchiveEvent(
                cid=dazl_event.cid)

        def wrap_method(func):
            handler = self._with_deferral(
                with_marshalling(
                    to_int_event,
                    without_return_value(
                        as_handler_invocation(
                            self.client, handler_status, func))))

            self.client.add_ledger_archived(template, match=match, handler=handler)

            return handler

        return wrap_method

    def get_status(self) -> 'IntegrationLedgerStatus':
        return IntegrationLedgerStatus(
            pending_calls=self.pending_queue.qsize(),
            sweep_events=self.sweep_events,
            event_handlers=self.handlers)

    async def start(self):
        return asyncio.create_task(self.worker())



import asyncio
import sys
import yaml

from dataclasses import dataclass
from datetime import datetime
from functools import wraps
from importlib import import_module
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Type
from zipfile import ZipFile

from dazl import AIOPartyClient, Network, Command
from dazl.util.prim_types import to_boolean

from dacite import from_dict, Config

from daml_dit_api import \
    DABL_META_NAME, \
    IntegrationEntryPoint, \
    IntegrationEnvironment, \
    IntegrationEvents, \
    IntegrationSpec, \
    IntegrationTimeEvents, \
    IntegrationTypeInfo, \
    IntegrationWebhookRoutes, \
    METADATA_INTEGRATION_ENABLED, \
    METADATA_INTEGRATION_RUN_AS_PARTY, \
    METADATA_INTEGRATION_TYPE_ID, \
    PackageMetadata

from .integration_time_context import \
    IntegrationTimeContext, IntegrationTimeStatus

from .integration_webhook_context import \
    IntegrationWebhookContext, IntegrationWebhookStatus

from .integration_ledger_context import \
    IntegrationLedgerContext, IntegrationLedgerStatus

from .common import \
    InvocationStatus, \
    without_return_value, \
    with_marshalling, \
    as_handler_invocation


from .log import LOG


@dataclass(frozen=True)
class IntegrationStatus:
    running: bool
    start_time: datetime
    error_message: 'Optional[str]'
    error_time: 'Optional[datetime]'
    webhooks: 'Optional[IntegrationWebhookStatus]'
    ledger: 'Optional[IntegrationLedgerStatus]'
    timers: 'Optional[IntegrationTimeStatus]'


def normalize_template_filter(template_filter: str):
    delim_loc = template_filter.find(':')

    if delim_loc >= 0:
        template_filter = template_filter[delim_loc+1:]

    return template_filter.replace(':', '.')


def normalize_metadata_field(field_value, field_type_info):
    LOG.debug('Normalizing %r with field_type_info: %r',
              field_value, field_type_info)

    field_value = field_value.strip()

    if field_type_info is not None \
       and field_type_info.field_type == 'contract_template':
        field_value = normalize_template_filter(field_value)

    return field_value.strip()


def normalize_metadata(metadata, integration_type):
    LOG.debug('Normalizing metadata %r for integration type: %r',
              metadata, integration_type)

    field_types = {field.id: field for field in integration_type.fields}

    return {field_id: normalize_metadata_field(field_value, field_types.get(field_id))
            for (field_id, field_value)
            in metadata.items()}


def _as_int(value: Any) -> int:
    return int(value)


def get_local_dabl_meta() -> 'Optional[str]':
    filename = f'pkg/{DABL_META_NAME}'

    LOG.info('Attmpting to load DABL metadata from local file: %r', filename)

    try:
        with open(filename, "r") as f:
            return f.read()
    except:  # noqa
        LOG.exception(f'Failed to local dabl meta {filename}')
        return None


def get_pex_dabl_meta() -> 'Optional[str]':
    pex_filename = sys.argv[0]

    LOG.info('Attmpting to load DABL metadata from PEX file: %r', pex_filename)

    try:
        with ZipFile(pex_filename) as zf:
            with zf.open(DABL_META_NAME) as meta_file:
                return meta_file.read().decode('UTF-8')
    except:  # noqa
        LOG.exception(f'Failed to read {DABL_META_NAME} from PEX file {pex_filename}'
                      f' (This is an expected error in local development scenarios.)')
        return None


def get_dabl_meta() -> str:
    meta = get_pex_dabl_meta() or get_local_dabl_meta()

    if meta:
        return meta

    raise Exception(f'Could not find {DABL_META_NAME}')


def parse_qualified_symbol(symbol_text: str):

    try:
        (module_name, sym_name) = symbol_text.split(':')
    except ValueError:
        raise Exception(f'Malformed symbol {symbol_text} (Must be [module_name:symbol_name])')

    module = None

    try:
        LOG.info(f'Searching for module {module_name} in qualified symbol {symbol_text}')
        module = import_module(module_name)
    except:  # noqa
        LOG.exception(f'Failure importing integration package: {module_name}')

    if module is None:
        raise Exception(f'Unknown module {module_name} in {symbol_text}')

    return (module, sym_name)


class IntegrationContext:

    def __init__(self,
                 network: 'Network',
                 integration_spec: 'IntegrationSpec'):

        iid = integration_spec.integration_id

        self.start_time = datetime.utcnow()

        self.network = network
        self.integration_spec = integration_spec
        self.dabl_meta = get_dabl_meta()

        self.running = False
        self.error_message = None  # type: Optional[str]
        self.error_time = None  # type: Optional[datetime]

        self.time_context = None  # type: Optional[IntegrationTimeContext]
        self.webhook_context = None  # type: Optional[IntegrationWebhookContext]
        self.ledger_context = None  # type: Optional[IntegrationLedgerContext]
        self.int_toplevel_coro = None

    def get_integration_entrypoint(
            self,
            integration_type: 'IntegrationTypeInfo') -> 'IntegrationEntryPoint':

        (module, entry_fn_name) = parse_qualified_symbol(integration_type.entrypoint)

        return getattr(module, entry_fn_name)

    def get_integration_env_class(
            self,
            integration_type: 'IntegrationTypeInfo') -> 'Type[IntegrationEnvironment]':

        if integration_type.env_class:
            (module, env_class_name) = parse_qualified_symbol(integration_type.env_class)

            return getattr(module, env_class_name)
        else:
            return IntegrationEnvironment

    def _get_integration_types(self):
        package_meta = from_dict(
            data_class=PackageMetadata, data=yaml.safe_load(self.dabl_meta))

        package_itypes = (package_meta.integration_types
                          or package_meta.integrations  # support for deprecated
                          or [])

        return {itype.id: itype for itype in package_itypes}

    async def _load_and_start(self):
        metadata = self.integration_spec.metadata
        LOG.info('=== REGISTERING INTEGRATION: %r', self.integration_spec)

        run_as_party = metadata[METADATA_INTEGRATION_RUN_AS_PARTY]
        client = self.network.aio_party(run_as_party)

        artifact_hash = self.integration_spec.artifact_hash

        integration_types = self._get_integration_types()

        integration_type = integration_types[self.integration_spec.type_id]

        env_class = self.get_integration_env_class(integration_type)
        entry_fn = self.get_integration_entrypoint(integration_type)

        metadata = normalize_metadata(metadata, integration_type)

        LOG.info("Starting integration with metadata: %r", metadata)

        integration_env_data = {
            **metadata,
            'party': run_as_party
            }

        integration_env = from_dict(
            data_class=env_class,
            data=integration_env_data,
            config=Config(type_hooks={
                bool: to_boolean,
                int: _as_int
            })
        )

        self.time_context = IntegrationTimeContext(client)
        self.ledger_context = IntegrationLedgerContext(client)
        self.webhook_context = IntegrationWebhookContext(client)

        events = IntegrationEvents(
            time=self.time_context,
            ledger=self.ledger_context,
            webhook=self.webhook_context)

        user_coro = entry_fn(integration_env, events)

        LOG.info("Waiting for ledger client to become ready")
        await client.ready()

        await self.ledger_context.process_sweeps()

        self.running = True
        LOG.info("Integration ready")

        int_coros = [
            self.time_context.start(),
            self.ledger_context.start()
        ]

        if user_coro:
            int_coros.append(user_coro)

        self.int_toplevel_coro = asyncio.gather(*int_coros)

    def get_status(self) -> 'IntegrationStatus':
        return IntegrationStatus(
            running=self.running,
            start_time=self.start_time,
            error_message=self.error_message,
            error_time=self.error_time,
            webhooks=self.webhook_context.get_status() if self.webhook_context else None,
            ledger=self.ledger_context.get_status() if self.ledger_context else None,
            timers=self.time_context.get_status() if self.time_context else None)

    async def safe_load_and_start(self):
        try:
            await self._load_and_start()
        except:  # noqa
            ex = sys.exc_info()[1]

            self.error_message = f'{repr(ex)} - {str(ex)}'
            self.error_time = datetime.utcnow()

            LOG.exception("Failure starting integration.")

    def get_coro(self):
        return self.int_toplevel_coro



import asyncio

from dataclasses import dataclass
from typing import List, Tuple, Optional, Callable, Sequence

from dazl import AIOPartyClient, Command

from .log import LOG

from daml_dit_api import \
    IntegrationTimeEvents

from .common import \
    InvocationStatus, \
    without_return_value, \
    as_handler_invocation


@dataclass
class TimerEventStatus(InvocationStatus):
    label: 'Optional[str]'


@dataclass
class IntegrationTimeStatus:
    timers: 'Sequence[TimerEventStatus]'


IntegrationTimerHandler = Callable[[], Sequence[Command]]


class IntegrationTimeContext(IntegrationTimeEvents):

    def __init__(self, client: 'AIOPartyClient'):
        self.queue = \
            asyncio.Queue(maxsize=1)  # type: asyncio.Queue[IntegrationTimerHandler]
        self.client = client
        self.intervals = []  # type: List[Tuple[int, IntegrationTimerHandler, TimerEventStatus]]

    def periodic_interval(self, seconds, label: 'Optional[str]' = None):
        def decorator(fn: 'IntegrationTimerHandler'):
            status = TimerEventStatus(
                index=len(self.intervals),
                label=label,
                command_count=0,
                use_count=0,
                error_count=0,
                error_message=None)

            wrapped = without_return_value(
                as_handler_invocation(
                    self.client, status, fn))

            self.intervals.append((seconds, wrapped, status))

            return wrapped
        return decorator

    async def worker(self):
        LOG.info('Time context worker starting...')
        while True:
            LOG.debug('...waiting for time event.')
            fn = await self.queue.get()
            try:
                LOG.debug('...received time event...')
                commands = await fn()

                if commands:
                    LOG.debug('Submitting ledger commands: %r', commands)
                    await self.client.submit(commands)

            except:  # noqa: E722
                LOG.exception('Uncaught error in time context worker loop')

    async def wait_loop(self, seconds, fn):
        LOG.info('Entering timer wait loop for %r second interval (fn: %r)',
                 seconds, fn)

        try:
            while True:
                LOG.debug('Wait loop waiting %r seconds...', seconds)
                await asyncio.sleep(seconds)
                await self.queue.put(fn)
        except:  # noqa: E722
            LOG.exception('Unexpected error in wait loop (%r, %r).', seconds, fn)

    async def start(self):
        worker_task = asyncio.create_task(self.worker())

        wait_tasks = [asyncio.create_task(self.wait_loop(seconds, fn))
                      for (seconds, fn, status)
                      in self.intervals]

        asyncio.gather(*[worker_task, *wait_tasks])

    def get_status(self) -> 'IntegrationTimeStatus':
        return IntegrationTimeStatus(
            timers=[status for (_, _, status) in self.intervals])



import os
from dataclasses import dataclass, asdict
from typing import Optional

from .log import LOG


@dataclass(frozen=True)
class Configuration:
    health_port: int
    ledger_url: str
    ledger_id: str
    integration_metadata_path: str


def env(var: str, default: 'Optional[str]' = None) -> str:
    val = os.getenv(var)

    LOG.debug('Configuration environment lookup: %r => %r (default: %r)', var, val, default)

    if val:
        return val
    elif default:
        return default
    else:
        raise Exception(f'Missing required environment variable: {var}')


def envint(var: str, default: 'Optional[int]' = None) -> int:
    val = env(var, str(default) if default else None)

    try:
        return int(val)
    except ValueError:
        raise Exception(f'Invalid integer {val} in environment variable: {var}')


def get_default_config() -> 'Configuration':
    config = Configuration(
        health_port=envint('DABL_HEALTH_PORT', 8089),
        ledger_url=env('DABL_LEDGER_URL', 'http://localhost:6865'),
        ledger_id=env('DABL_LEDGER_ID', 'cloudbox'),
        integration_metadata_path=env('DABL_INTEGRATION_METADATA_PATH', 'int_args.yaml'))

    LOG.info('Configuration: %r', asdict(config))

    return config



import asyncio
import base64
import logging
import sys
import yaml

from dataclasses import dataclass
from pathlib import Path
from asyncio import ensure_future, gather, get_event_loop
from typing import Optional, Dict
from dacite import from_dict

from aiohttp import ClientSession
from yarl import URL

from aiohttp.web import Application, AppRunner, TCPSite, RouteTableDef, \
    Request, Response

from dazl import Network

from daml_dit_api import IntegrationSpec

from .config import Configuration, get_default_config

from .log import LOG, setup_default_logging

from .web import start_web_endpoint

from .integration_context import IntegrationContext

import pkg_resources


def main():
    import logging

    setup_default_logging(level=logging.DEBUG)

    # Parsing certain DAML-LF modules causes very deep stacks; increase the standxrd limit
    # to be able to handle those.
    sys.setrecursionlimit(10000)

    loop = get_event_loop()
    loop.run_until_complete(_aio_main(get_default_config()))


def create_network(url: str) -> 'Network':
    network = Network()
    network.set_config(url=url)
    return network


def load_integration_spec(config: 'Configuration') -> 'IntegrationSpec':
    metadata_path = Path(config.integration_metadata_path)

    if metadata_path.exists():
        LOG.info('Loading integration metadata from: %r', metadata_path)

        return from_dict(
            data_class=IntegrationSpec,
            data=yaml.safe_load(metadata_path.read_bytes()))
    else:
        raise Exception('No metadata file found at: ' + repr(metadata_path))


async def _aio_main(config: 'Configuration'):
    LOG.info('Initializing dabl-integration...')

    network = create_network(config.ledger_url)
    dazl_coro = ensure_future(run_dazl_networks(network))

    integration_spec = load_integration_spec(config)

    integration_context = IntegrationContext(network, integration_spec)

    await integration_context.safe_load_and_start()

    web_coro = start_web_endpoint(config, integration_context)

    integration_coro = integration_context.get_coro()

    LOG.info('dabl-integration is now ready.')

    await gather(web_coro, dazl_coro, integration_coro)


async def run_dazl_networks(*networks: 'Network'):
    """
    Run the dazl networks, and make sure that fatal dazl errors terminate the application.
    """
    LOG.info('Starting dazl network(s)...')
    # noinspection PyBroadException
    try:
        futures = [ensure_future(n.aio_run()) for n in networks]
        await gather(*futures)
    except:  # noqa
        LOG.exception('The main dazl coroutine died with an exception')
        sys.exit(9)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['daml_dit_api', 'daml_dit_api.main']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp', 'dacite', 'dazl']

setup_kwargs = {
    'name': 'daml-dit-api',
    'version': '0.0.3',
    'description': 'DABL Integrations API Package',
    'long_description': None,
    'author': 'Mike Schaeffer',
    'author_email': 'mike.schaeffer@digitalasset.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/digital-asset/daml-dit-api',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import abc
import datetime

from dataclasses import asdict, dataclass, field
from typing import Any, Optional, Callable, Type, Sequence, TYPE_CHECKING, Mapping

from aiohttp.web import Response

from dazl import Command, ContractData, ContractId
from dazl.model.core import ContractMatch


def _empty_commands() -> 'Sequence[Command]':
    return list()


def _empty_tags() -> 'Sequence[str]':
    return list()


@dataclass(frozen=True)
class IntegrationTypeFieldInfo:
    id: str
    name: str
    description: str
    field_type: str
    help_url: 'Optional[str]' = None
    default_value: 'Optional[str]' = None
    required: 'Optional[bool]' = True


@dataclass(frozen=True)
class IntegrationTypeInfo:
    artifact_hash: 'Optional[str]'
    id: str
    name: str
    description: str
    fields: 'Sequence[IntegrationTypeFieldInfo]'
    entrypoint: str
    env_class: 'Optional[str]'
    runtime: 'Optional[str]' = 'python-file'
    help_url: 'Optional[str]' = None


@dataclass(frozen=True)
class CatalogInfo:
    name: str
    version: str
    description: str
    release_date: 'Optional[datetime.date]'
    author: 'Optional[str]'
    url: 'Optional[str]'
    email: 'Optional[str]'
    license: 'Optional[str]'
    experimental: 'Optional[bool]'
    demo_url: 'Optional[str]'
    source_url: 'Optional[str]'
    tags: 'Sequence[str]' = field(default_factory=_empty_tags)


DABL_META_NAME = 'dabl-meta.yaml'


@dataclass(frozen=True)
class PackageMetadata:
    catalog: 'Optional[CatalogInfo]'
    subdeployments: 'Optional[Sequence[str]]'
    integration_types: 'Optional[Sequence[IntegrationTypeInfo]]'

    # Deprecated in favor of integration_types
    integrations: 'Optional[Sequence[IntegrationTypeInfo]]'


@dataclass(frozen=True)
class IntegrationResponse:
    """
    Response to an integration event. All such responses contain
    a sequence of zero or more ledger commands to issue.
    """
    commands: 'Optional[Sequence[Command]]' = field(default_factory=_empty_commands)


class IntegrationTimeEvents:

    @abc.abstractmethod
    def periodic_interval(self, seconds, label: 'Optional[str]' = None):
        """
        Register a function as a handler for periodic timer events. The
        function will be scheduled to run at the specified
        interval. Note that while multiple interval timers can be used
        by a single integration, only one will ever run at once, with
        jobs skipped in the event of overruns.
        """
        pass


@dataclass(frozen=True)
class IntegrationLedgerContractEvent:
    """
    Base class for ledger integration events related to actions taken
    on contracts.
    """
    cid: ContractId


@dataclass(frozen=True)
class IntegrationLedgerContractCreateEvent(IntegrationLedgerContractEvent):
    """
    Event raised to notify an integration of a contract on ledger.  When
    an integration is started up, this event is raised for each contract
    of interest already on the ledger. While an integration is running,
    this event is raised within transaction boundaries to describe contracts
    created within that transaction. The extent of the contract's validity
    in the integration's event stream spans this event and the corresponding
    :class:`IntegrationLedgerContractArchiveEvent` for the CID.
    """
    initial: bool
    cdata: ContractData


@dataclass(frozen=True)
class IntegrationLedgerContractArchiveEvent(IntegrationLedgerContractEvent):
    """
    Event raised to notify an integration of a contract archive on ledger. This
    represents the end of the validity of the contact with this CID, and this
    event will never be raised before the integration receives a
    :class:`IntegrationLedgerContractCreateEvent` for the contract.
    """


@dataclass(frozen=True)
class IntegrationLedgerTransactionEvent:
    """
    Base class for transaction boundary events.
    """
    command_id: str
    workflow_id: str
    contract_events: 'Sequence[IntegrationLedgerContractEvent]'


@dataclass(frozen=True)
class IntegrationLedgerTransactionStartEvent(IntegrationLedgerTransactionEvent):
    """
    Event raised to signal the beginning of a ledger transaction. Outside of
    integration startup, all contract events occur between this and
    a matching transaction end event.
    """


@dataclass(frozen=True)
class IntegrationLedgerTransactionEndEvent(IntegrationLedgerTransactionEvent):
    """
    Event raised to signal the end of a ledger transaction. Outside of
    integration startup, all contract events occur between this and
    a matching transaction start event occurring earlier in the event stream.
    """


class IntegrationLedgerEvents:

    @abc.abstractmethod
    def ledger_init(self):
        """
        Decorator for registering a callback to be invoked when the
        ledger event stream has been connected and is initializing.
        Called before any :meth:`ledger_create` event, sweep or flow.
        """

    @abc.abstractmethod
    def ledger_ready(self):
        """
        Decorator for registering a callback to be invoked when the
        ledger event stream has been initialized. Called after
        any :meth:`ledger_create` event resulting from an initial sweep
        of contracs and before any flow events.
        """

    @abc.abstractmethod
    def transaction_start(self):
        """
        Decorator for registering a callback to be invoked when the
        integration receives a new transaction. Called before
        individual :meth:`ledger_create` and :meth:`ledger_archive`
        callbacks occurring within the transaction.
        """

    @abc.abstractmethod
    def transaction_end(self):
        """
        Decorator for registering a callback to be invoked when the
        integration receives a new transaction. Called after
        individual :meth:`ledger_create` and :meth:`ledger_archive`
        callbacks occurring within the transaction.
        """

    @abc.abstractmethod
    def contract_created(self, template: Any, match: 'Optional[ContractMatch]' = None,
                         sweep: bool = True, flow: bool = True):
        """
        Register a callback to be invoked when the integration encounters a newly created
        contract instance of a template.

        :param template:
            A template name to subscribe to, or '*' to subscribe on all templates.
        :param handler:
            The callback to invoke whenever a matching template is created.
        :param match:
            An (optional) parameter that filters the templates to be received by the callback.
        :param sweep:
            An (optional) parameter that controls whether or not the integration receives
            a sweep of all existing active contracts on startup.
        :param flow:
            An (optional) parameter that controls whether or not the integration receives
            contract messages after the initial sweep. (If both sweep and flow are false,
            the handler will never be called.)
        """

    @abc.abstractmethod
    def contract_archived(self, template: Any, match: 'Optional[ContractMatch]' = None):
        """
        Decorator for registering a callback to be invoked when the integration
        encounters a newly archived contract instance of a template.

        :param template:
            A template name to subscribe to, or '*' to subscribe on all templates.
        :param match:
            An (optional) parameter that filters the templates to be received by the callback.
        """


@dataclass(frozen=True)
class IntegrationWebhookResponse(IntegrationResponse):
    """
    Response to a webhook request. Contains both the
    ledger commands needed to fulfill the hook and the
    HTTP response.
    """

    response: 'Optional[Response]' = None


class IntegrationWebhookRoutes:

    @abc.abstractmethod
    def post(self, url_suffix: 'Optional[str]' = None, label: 'Optional[str]' = None):
        """
        Register a function as an HTTP POST handler for the integration's webhook.
        Integration HTTP handlers must return an instance of
        :class:`IntegrationWebhookResponse`.

        :param url_suffix:
            The suffix for the URL of the webhook resource. This is appended to
            an integration URL provided by DABL to form the full URL of the resource.

        :param label:
            A user-friendly description of the purpose of the webhook resource. This
            is displayed as on the DABL console to identify the purpose of the URL.
        """

    @abc.abstractmethod
    def get(self, url_suffix: 'Optional[str]' = None, label: 'Optional[str]' = None):
        """
        Register a function as an HTTP GET handler for the integration's webhook.
        Integration HTTP handlers must return an instance of
        :class:`IntegrationWebhookResponse`.

        :param url_suffix:
            The suffix for the URL of the webhook resource. This is appended to
            an integration URL provided by DABL to form the full URL of the resource.

        :param label:
            A user-friendly description of the purpose of the webhook resource. This
            is displayed as on the DABL console to identify the purpose of the URL.
        """


@dataclass
class IntegrationEvents:
    time: 'IntegrationTimeEvents'
    ledger: 'IntegrationLedgerEvents'
    webhook: 'IntegrationWebhookRoutes'


@dataclass
class IntegrationEnvironment:
    party: str


IntegrationEntryPoint = \
    Callable[[IntegrationEnvironment, IntegrationEvents], None]


METADATA_INTEGRATION_ID = 'com.projectdabl.integrations.integrationId'
METADATA_INTEGRATION_TYPE_ID = 'com.projectdabl.integrations.integrationTypeId'
METADATA_INTEGRATION_COMMENT = 'com.projectdabl.integrations.comment'
METADATA_INTEGRATION_ENABLED = 'com.projectdabl.integrations.enabled'
METADATA_INTEGRATION_RUN_AS_PARTY = 'com.projectdabl.integrations.runAsParty'
METADATA_INTEGRATION_RUNTIME = 'com.projectdabl.integrations.runtime'


@dataclass(frozen=True)
class IntegrationSpec:
    integration_id: str
    artifact_hash: str
    type_id: str
    enabled: bool
    metadata: 'Mapping[str, str]'
    runtime: 'Optional[str]' = 'python-file'

    @classmethod
    def from_dict(cls, json_obj):
        integration_id = json_obj.get('integration_id')
        artifact_hash = json_obj.get('artifact_hash')
        type_id = json_obj.get('type_id')
        enabled = json_obj.get('enabled')
        metadata = json_obj.get('metadata')
        runtime = json_obj.get('runtime')

        return cls(integration_id=integration_id,
                   artifact_hash=artifact_hash,
                   type_id=type_id,
                   enabled=enabled,
                   metadata=metadata,
                   runtime=runtime)

    def to_dict(self) -> 'Mapping[str, Any]':
        return asdict(self)


def is_true(value: 'Optional[str]') -> bool:
    if value is None:
        return False

    return value.lower() in ['1', 'yes', 'true', 'y', 't']
